package registry

import (
	"context"
	"encoding/json"
	"fmt"
	"github.com/odysseus/service-registry/common"
	"github.com/redis/go-redis/v9"
	log "github.com/sirupsen/logrus"
	"sync"
	"time"
)

type RegistryInfo struct {
	Timestamp int64  `redis:"timestamp" json:"timestamp"`
	Instance  string `redis:"instance" json:"instance"` // example: hostip + hostname
	Status    string `redis:"status" json:"status"`     // example: kafka failed or running.
	Detail    string `redis:"detail" json:"detail"`
}

func (g RegistryInfo) TimeStamp() int64 {
	return g.Timestamp
}

func (g RegistryInfo) Message() json.RawMessage {
	d, _ := json.Marshal(g)
	return d
}

type Register interface {
	ServiceType() common.ServiceType
	Instance() string
	Status() string
	DetailInfo() (json.RawMessage, error)
}

type Registry struct {
	rdb      *redis.Client
	rw       sync.RWMutex
	register Register
	quit     chan struct{}
}

type RedisConnParam struct {
	Addr     string
	Password string
	DbIndex  int
}

func NewRegistry(redisParam RedisConnParam, register Register) *Registry {
	switch register.ServiceType() {
	case common.SERVICE_NODE_MANAGER, common.SERVICE_API_GATEWAY, common.SERVICE_BACKEND, common.SERVICE_SCHEDULER, common.SERVICE_WORKER:
		//nothing
	default:
		log.WithField("service", register.ServiceType()).Error("not support service type")
		return nil
	}
	rdb := redis.NewClient(&redis.Options{
		Addr:     redisParam.Addr,
		Password: redisParam.Password,
		DB:       redisParam.DbIndex,
	})

	return &Registry{
		rdb:      rdb,
		register: register,
		quit:     make(chan struct{}),
	}
}

func (s *Registry) Start() {
	ticker := time.NewTicker(time.Second * 1)
	defer ticker.Stop()

	for {
		select {
		case <-s.quit:
			return
		case <-ticker.C:
			if err := s.registry(s.rdb); err != nil {
				log.WithError(err).Error("registry failed")
			} else {
				ticker.Reset(time.Second * 10)
			}
		}
	}
}

func (s *Registry) Stop() {
	close(s.quit)
}

func (s *Registry) registry(rdb *redis.Client) error {
	k := fmt.Sprintf("%s%s", common.GetServiceKeyPrefix(s.register.ServiceType()), s.register.Instance())
	detail, err := s.register.DetailInfo()
	if err != nil {
		log.WithError(err).Error("get detail info failed")
		return err
	}
	status := s.register.Status()

	err = rdb.HSet(context.Background(), k, RegistryInfo{
		Timestamp: time.Now().Unix(),
		Status:    status,
		Instance:  s.register.Instance(),
		Detail:    string(detail),
	}).Err()
	if err != nil {
		log.WithError(err).Error("set register info failed")
	}
	return err
}
